#include "pdc2.h"

void Periodic_Smooth(SCALAR **, int, int);
void Digital_Smooth(SCALAR **, int, int);
int realft(SCALAR *,int,int);
/* Corrections /additions by P.G. */

/***************************************************************/
void gather(isp)
int isp;
{
  register int i, j,rj,thj,n;
  register SCALAR RI,TI,denom,rpj,rmj,kmt; 
  SCALAR    temp;
        /* These could be #define's but if they can be registers it would be 
           much faster */


  /********************************************************************/
  /* Step 1: */
  /* Start of Gather */

    for (i=0; i< ngr; i++)
      for (j=0; j< ngt; j++) {
	srho0[isp][i][j] = srho1[isp][i][j]+sp_n_mcc[isp][i][j];
	srho1[isp][i][j] = sp_n_mcc[isp][i][j] = 0;
      }
        
/* weight the particles to the grid using a Cloud-in-Cell algorithm  */

    for (n=np[isp]-1; n>=0; n--) {
      
      /* rj & thj are ints, not SCALARs!! */
      /* RI and TI are SCALARs!! */

/*   since grid 0 is inner conductor, have to account for that */

      rj = (int) (r[isp][n] - rminn); /* rminn = 0 when rmin = 0.0 */
      thj = th[isp][n];

      RI = r[isp][n]-rj-rminn;
      TI = th[isp][n]-thj;

      if(rj) {            /* particle is not in first r cell, use CIC */
	/* remember that r[isp][n] is between 0 and ngr-1 and th[isp][n] is 
           between 0 and ngt-1 w1,w2,w3,w4 are the weight to the grid points, 
           they are being saved for scatter() This new weighting scheme runs 
           like gangbusters. It is saving 10 operations per particle per time 
           step! */
	srho1[isp][rj][thj]    +=(w1[isp][n]=(1-RI)*(1-TI));
	srho1[isp][rj+1][thj]  +=(w2[isp][n]=RI*(1-TI));
	srho1[isp][rj][thj+1]  +=(w3[isp][n]=(1-RI)*TI);
	srho1[isp][rj+1][thj+1]+=(w4[isp][n]=RI*TI);
      }
      else { 

        if (cent_cond)
        {
           srho1[isp][rj][thj]    +=(w1[isp][n]=(1-RI)*(1-TI));
           srho1[isp][rj+1][thj]  +=(w2[isp][n]=RI*(1-TI));
           srho1[isp][rj][thj+1]  +=(w3[isp][n]=(1-RI)*TI);
           srho1[isp][rj+1][thj+1]+=(w4[isp][n]=RI*TI);
        }
        else
        {
	/* rj = 0, particle is in first r cell, use PIC 
	   Remember that the first cell is just a sector,
	   not a rectangular-like area 
	   */
	
	/* these expressions can be reduced */
	  srho1[isp][1][thj]   += (w2[isp][n] = RI*(1-TI));
	  srho1[isp][1][thj+1] += (w3[isp][n] = RI*TI);
	  srho1[isp][0][0]     += (w1[isp][n] = 1-RI);
	}
      }
    }
	    
    /* Weight the end cells properly */
    /* Basically, take into account the fact that the end cells 
       get weights from only one side */
/*    for (j=0; j< ngt; j++) {
      srho1[isp][ncr][j] *= 2;
    }
*/
    /* Accumulate charges on the real origin to the polar origin */
    if (!cent_cond)
    {
      for(j=1;j < ngt;j++) srho1[isp][0][j] = srho1[isp][0][0];
    }
    else
    {
      srho1[isp][0][0] = srho1[isp][0][nct] = srho1[isp][0][0]
                                              + srho1[isp][0][nct];	
    }
    /* enforce a periodic boundary condition around theta */
    for (i=1; i< ngr; i++)
      srho1[isp][i][0] = srho1[isp][i][nct] = srho1[isp][i][0] 
                                              + srho1[isp][i][nct];

    for(i=0;i < ngr; i++)
      for(j=0;j < ngt; j++)
          srho1[isp][i][j] /= vol[i];

    /* ---------------- leehj ----------------*/
    for (i=0;i<nsmooth; i++) Digital_Smooth(srho1[isp], ncr, nct);
    if(!cent_cond && BCflag) {
      for(j=0;j<nct/2;j++)
        srho1[isp][0][j] = srho1[isp][0][j+nct/2]
                         = (srho1[isp][0][j] + srho1[isp][0][j+nct/2])/2.;
    }
}    

  /********************************************************************/

void fields()
{
  register int i, j, k, isp;
  register SCALAR bet, s, sn, cs;
  SCALAR jm2, jm1, jo, jp1, jp2,tmp;

  /* Convert number per unit volume(srho) into Charge per unit volume(rho) */
    for (i=0; i< ngr; i++)
      for (j=0; j< ngt; j++) rho[i][j]=dust_rho[i][j];
  for (isp=0;isp<nsp;isp++) 
    for (i=0; i< ngr; i++)
      for (j=0; j< ngt; j++) 
        rho[i][j] += srho[isp][i][j]*q[isp]/drdthzlength;

  /* Step 2 */
  /* Fourier charge density in theta */
  /* Transform theta for a constant r */

  /* Pass rhok[i]-1 because fft routine uses arrays that start from 1 */
  for (i=0; i< ngr; i++) {
    for (j=0; j< ngt; j++) 
      rhok[i][j] = rho[i][j];
    realft(rhok[i]-1, fftngt, 1);
    rhok[i][1] = 0.0; /* smooth rho */
  }
  if (bias_grid>0 && bias_grid<ngr) {
   rhok[bias_grid][0]=bias_volt/se;
   for (k=1;k<ngt;k++) rhok[bias_grid][k]=0;
  }
  /**************************************************************************/

  /* Step 3 */
  /* Solve a tridiagonal matrix to get phik[i][k] 
     This is a rewritten slightly TRIDAG from Numerical Recipes 
     Don't forget the Jack Handey rule: FREE DUMMY!! 
     phik[0][m>0]=0 because a wave can't propagate on only one grid point
     phik[ngr-1][m>0]=0  because the wall is metal */

  /* Solve a tridiagonal matrix to solve the dc case, including the origin and the wall */

/*-----------------------------------*/
/* Solve for phik[0][...] in k space */
/*-----------------------------------*/
  phik[ncr][0]= outer_dc; /* Set the wall potential */

  if (cent_cond) {
    jp = backj;
    for (isp=0; isp< nsp; isp++) jp += jwall[isp];
    if(ntri != 1) phik[0][0] = inner_dc;
    phik[ntri-1][0] = (*bcptr)()/b0;
    bet=B[1][0]; /* hoo baby */
    if (BCflag) { phik[0][0]=0; ntri=1; bet=B[0][0];}
  }
  else {
    phik[0][0] = (pidrdr*rhok[0][0]/(nct*epsdth2))/(bet=B[0][0]);
    ntri = 1;
  }
 
  for(i=ntri;i < ncr; i++) {
    gam[i] = C[i-1]/bet;
    bet = B[i][0] - A[i]*gam[i];
    phik[i][0] = (se*rhok[i][0] -A[i]*phik[i-1][0])/bet;
    if (i==ncr-1) phik[i][0]-= C[i]*phik[ncr][0]/bet;
  }
  for(i=ncr-2;i >= ntri-1 ;i--)
    phik[i][0] -= gam[i+1]*phik[i+1][0];
 
/*--------------------------------------------------------------------*/
/* Solve a tridiagnal matrix to get the ac part of phik[i][k], k >= 1 */  
/*--------------------------------------------------------------------*/

  for (k=1; k< ngt; k++)
    phik[0][k] = phik[ncr][k] = 0.0; 
                            /* Can't propagate a wave on origin or wall */
   /* leehj : Change the wall potential as a rotating field */
  if (outer_mode && t>outer_t0) {
    phik[ncr][2*outer_mode]  =-outer_ac*sin(outer_f0*t-delta2);
    phik[ncr][2*outer_mode+1]= outer_ac*cos(outer_f0*t-delta2);
  }
  if (cent_cond && inner_mode && t>inner_t0 && !BCflag) {
    phik[0][2*inner_mode]  =-inner_ac*sin(inner_f0*t-delta1);
    phik[0][2*inner_mode+1]= inner_ac*cos(inner_f0*t-delta1);
  }

 if (BCflag==0) {  
  for (k=1; k< nct; k++) {
    phik[1][k] = (se*rhok[1][k]-A[1]*phik[0][k])/(bet=B[1][k]);
    for(i=2; i < ncr; i++) {
      gam[i] = C[i-1]/bet;
      bet = B[i][k] -A[i]*gam[i];
      phik[i][k] = (se*rhok[i][k] - A[i]*phik[i-1][k])/bet;
      if (i==ncr-1) phik[i][k]-= C[i]*phik[ncr][k]/bet; 
    }
    for(i=ncr-2; i>=1; i--) 
      phik[i][k] -= gam[i+1]*phik[i+1][k];
  }
 } 
 else {
  for (k=1; k< nct; k++) {
    bet=B[0][k]; 
    for (i=1; i<ncr; i++) {
      gam[i] = C[i-1]/bet;
      bet = B[i][k] -A[i]*gam[i];
      phik[i][k] = (se*rhok[i][k] -A[i]*phik[i-1][k])/bet;
      if (i==ncr-1) phik[i][k]-=C[i]*phik[ncr][k]/bet;
    }
    for(i=ncr-2; i>=0; i--)
      phik[i][k] -= gam[i+1]*phik[i+1][k];
  }
 } 
 
  /********************************************************************/
  /* Step 4 */
  /* Take inverse Fourier Transform of phik */

  for (i=0; i< ngr; i++) {
    for (j=0; j< nct; j++)  /* pdp2 has nc not ng */
      phi[i][j] = phik[i][j];
    realft(phi[i]-1, fftngt, -1); 
     phi[i][nct] = phi[i][0]; 
  }
  

  /* Now really, really enforce a periodic boundary condition dammit */
/*  for(i=0;i<ngr;i++)
    phi[i][0] = phi[i][nct] = (phi[i][0] + phi[i][nct])/2;
*/
  /********************************************************************/

  /* Step 5 */

  /* calculate charge density on walls */
  if (cent_cond && !BCflag) (*circuitptr)();

/* Solve for E-field using a half-assed second-order finite difference scheme */

  s = 0.5/dr;
  for(j=0; j < ngt; j++) {
    er[ncr][j] = (phi[ncr-1][j] - phi[ncr][j])/dr + 
	           rho[ncr][j]*0.5*dr/epsilon;
    sigma_wall[j] = er[ncr][j]*epsilon;
    for (i=1; i< ncr; i++) 
      er[i][j] = (phi[i-1][j] - phi[i+1][j])*s;
  }
  /* No E-field in theta direction on wall */
  k=ncr;
  if (outer_mode) k=ngr; 
  else  for (j=0; j< ngt; j++) eth[ncr][j] = 0.0;  

  s = 0.5/dth;
  for(i=0; i < k; i++) {
    eth[i][nct] = eth[i][0] = (phi[i][nct-1] - phi[i][1])*s/r_array[i];
    for (j=1; j< nct; j++) 
      eth[i][j] = (phi[i][j-1] - phi[i][j+1])*s/r_array[i];
  } /* 4.5.93 should j=1;j<ngt-1 be j=0??? */

/* PG addition */
  if (cent_cond)
  {
    s = 1.0/dr;
    for(j=0;j<ngt; j++) { 
      er[0][j] = (rminn+0.5)*(phi[0][j]-phi[1][j])*s/rminn;
      if (!BCflag) er[0][j]-=(rminn+0.25)*rho[0][j]*0.5*dr/epsilon/rminn;
    }
  }
  else
  {
  /* Solve for the E-field at the origin by using e = - grad phi 
     But you have you keep the vector quantities correct.
     Also the E-field at the origin must in X-Y. 
     PM 1/27/93 */
    ex=ey=0;
    s=0.5/dr;
    for(j=0;j< nct/2; j++) {
      ex += s*(phi[1][nct/2+j]-phi[1][j])*cos_array[j];
      ey += s*(phi[1][nct/2+j]-phi[1][j])*sin_array[j];
    }
    ex*=2.0/nct;
    ey*=2.0/nct;
    /* djc addition 
    for(j=0;j<nct; j++) {
      er[0][j] = ex*cos_array[j] + ey*sin_array[j];
      eth[0][j] = ey*cos_array[j] - ex*sin_array[j];
    }
    er[0][nct]=er[0][0];
    eth[0][nct]=eth[0][0];
    ------*/
    /*-------- HJLEE addition ----------
    for(j=0;j<ngt; j++) {
      eth[0][j] = 0;
      er[0][j] = 0;
    }
    s=sqrt(ex*ex+ey*ey);
    tmp=0.5*acos(ex/s)/M_PI*nct;
    i=(int)tmp;
    if (ey<0) i+=nct/2;
    er[0][i]=s*(i+1-tmp);
    er[0][i+1]=s*(tmp-i);
    ----------------*/
    for(j=0;j<ngt; j++) {
      eth[0][j] = 2*eth[1][j]-eth[2][j];
      er[0][j] = 2*er[1][j]-er[2][j];
    }
  }

  for (i=0;i<hngr;i++) 
    for (j=0;j<hngt;j++) {
      er2[j][i]=er[i*ngfactor_r][j*ngfactor_t];
      eth2[j][i]=eth[i*ngfactor_r][j*ngfactor_t];
    }
}

/***************************************************************/
/*                               1 2 1                          */
/* Smoothing the array using the 2 4 2 method with the proper  */
/*                               1 2 1                          */
/* boundary conditions to conserve charge.                     */

void Periodic_Smooth(SCALAR **a, int nx, int ny)
{
  register int i,j;
  static int init_flag = 1;
  static SCALAR **temp;
  SCALAR tmp;

  if (init_flag) {
    temp= (SCALAR **)malloc((nx+1)*sizeof(SCALAR *));
    for(i=0; i<=nx; i++) {
      if(!(temp[i]= (SCALAR *)malloc((ny+1)*sizeof(SCALAR)))) {
        puts("null ptr in 1-4-1");
        exit(1);
      }
    }
    init_flag=0;
  }
  
  for (j=1; j<ny; j++) {
    temp[0][j] = ( 2.0*a[0][j] + a[0][j+1] + a[0][j-1] + 2.0*a[1][j] + a[1][j+1]
               + a[1][j-1])/8.0;
    temp[nx][j] = ( 2.0*a[nx][j] + a[nx][j+1] + a[nx][j-1] + 2.0*a[nx-1][j]
                + a[nx-1][j+1] + a[nx-1][j-1])/8.0;
    for (i=1; i<nx; i++)
      temp[i][j] = (4.0*a[i][j] + 2.0*a[i][j+1] +2.0*a[i][j-1] + 2.0*a[i+1][j] 
                 + 2.0*a[i-1][j] + a[i+1][j+1] + a[i-1][j+1] + a[i+1][j-1] 
                 + a[i-1][j-1])/16.0;
  }

  temp[0][ny] = temp[0][0] = (2.0*a[0][0] + a[0][1] + a[0][ny-1] + 2.0*a[1][0]
                           + a[1][1] + a[1][ny-1])/8.0;
  temp[nx][ny] = temp[nx][0] = (2.0*a[nx][0] + a[nx][1] + a[nx][ny-1] 
                             + 2.0*a[nx-1][0] + a[nx-1][1] + a[nx-1][ny-1])/8.0;
  for (i=1; i<nx; i++)
    temp[i][ny] = temp[i][0] = (4.0*a[i][0] + 2.0*a[i][1] + 2.0*a[i][ny-1] 
               + 2.0*a[i+1][0] + 2.0*a[i-1][0] + a[i+1][1] + a[i+1][ny-1] 
               + a[i-1][1] + a[i-1][ny-1])/16.0;

/*  for (i=0;i<=nx;i++) temp[i][ny] = temp[i][0] = (temp[i][1]+temp[i][ny-1])/2;
*/
  for (i=0; i<=nx; i++)
    for (j=0; j<=ny; j++)  a[i][j]=temp[i][j];
  
  if (!cent_cond) {
    tmp=0;
    for (j=0; j<ny; j++) tmp += a[0][j];
    tmp/=ny;
    for (j=0; j<=ny; j++) a[0][j]=tmp;
  }
}                                                                                
void Digital_Smooth(SCALAR **a, int nr, int nt)
{
  register int i,j;
  static int init_flag = 1;
  static SCALAR **temp, *alpha;
  SCALAR tmp;

  if (init_flag) {
    temp= (SCALAR **)malloc((nr+1)*sizeof(SCALAR *));
    alpha=(SCALAR *)malloc((nr+1)*sizeof(SCALAR));	
    for(i=0; i<=nr; i++) {
      if(!(temp[i]= (SCALAR *)malloc((nt+1)*sizeof(SCALAR)))) {
        puts("null ptr in the filter");
        exit(1);
      }
    }
    init_flag=0;
  }
  
  alpha[0]=0.5;
  alpha[1]=alpha[0]*vol2[0]/vol2[1];
  for (j=2;j<nr;j++) 
    alpha[j]=(2*alpha[j-1]*vol2[j-1]-alpha[j-2]*vol2[j-2])/vol2[j];
  alpha[nr]=alpha[nr-1]*vol2[nr-1]/vol2[nr];

  for (j=0; j<=nt; j++) {
    temp[0][j] = 0.5*(a[0][j]+a[1][j]);
    temp[nr][j] = alpha[nr]*a[nr-1][j]+(1-alpha[nr])*a[nr][j];
    for (i=1; i<nr; i++)
      temp[i][j] = alpha[i]*(a[i-1][j]+a[i+1][j]) + (1-2*alpha[i])*a[i][j];
  }

 /* 
  for (j=0; j<=nt; j++) {
    a[0][j]=0.5*(temp[0][j]+temp[1][j]);
    a[nr][j]=alpha[nr]*temp[nr-1][j]+(1-alpha[nr])*temp[nr][j];
    for (i=1; i<nr; i++)
      a[i][j]=alpha[i]*(temp[i-1][j]+temp[i+1][j]) + (1-2*alpha[i])*temp[i][j];
  }

  for (j=0; j<=nt; j++) {
    temp[0][j] = 0.5*(a[0][j]+a[1][j]);
    temp[nr][j] = alpha[nr]*a[nr-1][j]+(1-alpha[nr])*a[nr][j];
    for (i=1; i<nr; i++)
      temp[i][j] = alpha[i]*(a[i-1][j]+a[i+1][j]) + (1-2*alpha[i])*a[i][j];
  }
*/
  for (i=0; i<=nr; i++) {
    for (j=1; j<nt; j++) a[i][j]=0.25*(temp[i][j-1]+2*temp[i][j]+temp[i][j+1]);
    a[i][0]=a[i][nt]=0.25*(temp[i][nt-1]+2*temp[i][0]+temp[i][1]);
  }

  if (!cent_cond) {
    tmp=0;
    for (j=0; j<nt; j++) tmp += a[0][j];
    tmp/=nt;
    for (j=0; j<=nt; j++) a[0][j]=tmp;
  }
}  
