#include "pdc2.h"
#include "xgrafix.h"

#define   NEMAX   500
#define   NIMAX   500

void hnewvel(SCALAR energy, SCALAR vel, SCALAR *vx, SCALAR *vy, SCALAR *vz, int e_flag);
SCALAR hsigma1(SCALAR), hsigma2(SCALAR), hsigma3(SCALAR), hsigma4(SCALAR);
SCALAR gden, vgth, extengy0, ionengy0;
int ecolsp, icolsp;
int maxwellv(SCALAR *, SCALAR *, SCALAR *, SCALAR);

/*********************************************************/
/* Monte Carlo Collisions for electron- and ion-neutrals */

void heliummcc(int isp)
{
  static int init_mcc_flag=1;
  static SCALAR ecol_extra, icol_extra;
  static SCALAR max_sigmav_e, max_sigmav_i;
  static SCALAR col_prob_e, col_prob_i;

  register int i, j, k, index;
  int N, nnp, s1, s2;      
  SCALAR    delx, dely;
  SCALAR random, vel, sigma_total;
  SCALAR engy, rengy, phi1;
  SCALAR cosphi, sinphi, coschi, sinchi;
  SCALAR sum_sigma, temp, vneutx, vneuty, vneutz;
  
  if (init_mcc_flag) {
    gden = NperTORR*pressure/(gtemp+DBL_MIN);  /* Calculating the gas density */
    
    ecolsp= ecollisional-1;
    icolsp= icollisional-1;
    ionsp--;
    if(ionsp>0) vgth= sqrt(2*1.602e-19*gtemp/m[ionsp]);
     
    /*****************************************/
    /* Calculating the null collision prob.  */
    
    if(ecollisional) {
      extengy0 = 19.8;
      ionengy0 = 24.5;
      
      max_sigmav_e = 0.0;
      for (engy=0; engy<NEMAX; engy += 0.1)
	max_sigmav_e= max(max_sigmav_e,
			  sqrt(2*1.602e-19*engy/m[ecolsp])
			  *(hsigma1(engy)+hsigma2(engy)+hsigma3(engy)));
      
      col_prob_e = 1 -exp(-max_sigmav_e*gden*dt*sp_k[ecolsp]);
      ecol_extra = 0.5;
    }
    if(icollisional) {
      max_sigmav_i = 0.0;
      for (engy=0; engy<NIMAX; engy += 0.1)
	max_sigmav_i= max(max_sigmav_i, 
                          sqrt(2*1.602e-19*engy/m[icolsp])*hsigma4(engy));
      
      col_prob_i = 1 -exp(-max_sigmav_i*gden*dt*sp_k[icolsp]);
      icol_extra = 0.5;
    }
    printf("e_col_prob=%e, i_col_prob=%e\n",col_prob_e,col_prob_i);
    init_mcc_flag = 0;
  }
   
  /********************************/
  /* Electron collisions with He  */
  
  if(ecollisional && isp==ecolsp) {
    if(theRunWithXFlag) {
      for (j=0; j<ngr; j++)
        for (k=0; k<ngt; k++)
          mccrate[0][j][k] = mccrate[1][j][k] = mccrate[2][j][k] = 0.0;
    }
    ecol_extra += np[ecolsp]*col_prob_e;
    N = ecol_extra;
    ecol_extra -= N;
    
    nnp = np[ecolsp];
    for(j=0; j< N; j++) {
      index= nnp*frand();
      nnp--;
      temp = r[ecolsp][nnp];
      r[ecolsp][nnp] = r[ecolsp][index];
      r[ecolsp][index] = temp;
      
      temp = th[ecolsp][nnp];
      th[ecolsp][nnp] = th[ecolsp][index];
      th[ecolsp][index] = temp;
      
      temp = vr[ecolsp][nnp];
      vr[ecolsp][nnp] = vr[ecolsp][index];
      vr[ecolsp][index] = temp;
      
      temp = vth[ecolsp][nnp];
      vth[ecolsp][nnp] = vth[ecolsp][index];
      vth[ecolsp][index] = temp;
      
      temp = vz[ecolsp][nnp];
      vz[ecolsp][nnp] = vz[ecolsp][index];
      vz[ecolsp][index] = temp;
    }
    
    for(j=nnp; j<nnp+N; j++) {
      engy = Exscale[ecolsp]*vr[ecolsp][j]*vr[ecolsp][j];
      engy+= Eyscale[ecolsp]*vth[ecolsp][j]*vth[ecolsp][j];
      engy+= Ezscale[ecolsp]*vz[ecolsp][j]*vz[ecolsp][j];
      vel  = sqrt(2.0*engy/m[ecolsp]);

      engy*=6.24219725e18;

      sigma_total = max_sigmav_e/(vel + DBL_MIN);   /**djc**/
      random= frand();
      
      /*********************************************************************/
      /* determine the type of collision and calculate new velocities, etc */
      
      /*******************************/
      /* if the collision is elastic */
      
      if (random <= (sum_sigma =hsigma1(engy))/sigma_total) { 
	/* first normalize vel */
	vr[ecolsp][j] *= vrscale[ecolsp]/vel;
	vth[ecolsp][j] *= vthscale[ecolsp]/vel;
	vz[ecolsp][j] *= vzscale[ecolsp]/vel;
	
	/* scatter the electron */
	hnewvel(engy, vel, &vr[ecolsp][j], &vth[ecolsp][j], &vz[ecolsp][j], 1);
	
	/* normalize back to deminsionless vel */
	vr[ecolsp][j] /= vrscale[ecolsp];
	vth[ecolsp][j] /= vthscale[ecolsp];
	vz[ecolsp][j] /= vzscale[ecolsp];
        /* take diagnostics */
        if (theRunWithXFlag) {
          s1 = r[ecolsp][j]-rminn;
          s2 = th[ecolsp][j];
          delx = r[ecolsp][j]-rminn -s1;
          dely = th[ecolsp][j] -s2;
          mccrate[0][s1][s2]    += (1-delx)*(1-dely);
          mccrate[0][s1+1][s2]  += delx*(1-dely);
          mccrate[0][s1][s2+1]  += (1-delx)*dely;
          mccrate[0][s1+1][s2+1]+= delx*dely;
        }
      }
      /**********************************/
      /* if the collision is excitation */
      
      else if (engy >= extengy0 && random <= (sum_sigma +=hsigma2(engy))/sigma_total) {
	/* first normalize vel */
	vr[ecolsp][j] *= vrscale[ecolsp]/vel;
	vth[ecolsp][j] *= vthscale[ecolsp]/vel;
	vz[ecolsp][j] *= vzscale[ecolsp]/vel;
	engy -= extengy0;
	vel   = sqrt(2.0*1.602e-19*engy/m[ecolsp]);
	
	/* scatter the electron */
	hnewvel(engy, vel, &vr[ecolsp][j], &vth[ecolsp][j], &vz[ecolsp][j], 0);
	/* normalize back to deminsionless vel */
	vr[ecolsp][j] /= vrscale[ecolsp];
	vth[ecolsp][j] /= vthscale[ecolsp];
	vz[ecolsp][j] /= vzscale[ecolsp];
        if (theRunWithXFlag) {
          s1 = r[ecolsp][j]-rminn;
          s2 = th[ecolsp][j];
          delx = r[ecolsp][j]-rminn -s1;
          dely = th[ecolsp][j] -s2;
          mccrate[1][s1][s2]    += (1-delx)*(1-dely);
          mccrate[1][s1+1][s2]  += delx*(1-dely);
          mccrate[1][s1][s2+1]  += (1-delx)*dely;
          mccrate[1][s1+1][s2+1]+= delx*dely;
        }
      } 
      /***************************************************************************/
      /* if the collision is ionization, add the released electron and ion first */
      
      else if(engy >= ionengy0 && random <= (sum_sigma +=hsigma3(engy))/sigma_total) {
	/* first normalize vel */
	vr[ecolsp][j] *= vrscale[ecolsp]/vel;
	vth[ecolsp][j] *= vthscale[ecolsp]/vel;
	vz[ecolsp][j] *= vzscale[ecolsp]/vel;
	
	/********************************/
	/* subtract the ion. energy and */
	/* partition the remaining energy */
	
	engy -= ionengy0;
	rengy = 10.0*tan(frand()*atan(engy/20.0));
	engy -= rengy;
	
	/********************************/
	/* scatter the created electron */
	
	vel= sqrt(2.0*1.602e-19*rengy/m[ecolsp]);
	k  = np[ecolsp];
	vr[ecolsp][k] = vr[ecolsp][j];
	vth[ecolsp][k] = vth[ecolsp][j];
	vz[ecolsp][k] = vz[ecolsp][j];
	hnewvel(rengy, vel, &vr[ecolsp][k], &vth[ecolsp][k], &vz[ecolsp][k], 0);
	/* normalize back to deminsionless vel */
	vr[ecolsp][k] /= vrscale[ecolsp];
	vth[ecolsp][k] /= vthscale[ecolsp];
	vz[ecolsp][k] /= vzscale[ecolsp];
	
	r[ecolsp][k] = r[ecolsp][j];
	th[ecolsp][k] = th[ecolsp][j];
	x[ecolsp][k]= r[ecolsp][k]*cos(th[ecolsp][k]*dth);
	y[ecolsp][k]= r[ecolsp][k]*sin(th[ecolsp][k]*dth);
	/****************************************/
	/* assign velocities to the created ion */
	
	k = np[ionsp];
	maxwellv(&vr[ionsp][k], &vth[ionsp][k], &vz[ionsp][k], vgth);
	vr[ionsp][k] /= vrscale[ionsp];
	vth[ionsp][k] /= vthscale[ionsp];
	vz[ionsp][k] /= vzscale[ionsp];
	
	r[ionsp][k] = r[ecolsp][j];
	th[ionsp][k] = th[ecolsp][j];
	/*****************************************/
	/* finally scatter the incident electron */
	
	vel= sqrt(2.0*1.602e-19*engy/m[ecolsp]);
	hnewvel(engy, vel, &vr[ecolsp][j], &vth[ecolsp][j], &vz[ecolsp][j], 0);
	
	/* normalize back to deminsionless vel */
	vr[ecolsp][j] /= vrscale[ecolsp];
	vth[ecolsp][j] /= vthscale[ecolsp];
	vz[ecolsp][j] /= vzscale[ecolsp];
	
	if(++np[ionsp] >maxnp[ionsp] || ++np[ecolsp] >maxnp[ecolsp]) {
	  puts("ADJUST(Ionization): too many particles. MUST EXIT!");
        }
	s1 = r[ecolsp][j]-rminn;
	s2 = th[ecolsp][j];
	delx = r[ecolsp][j]-rminn -s1;
	dely = th[ecolsp][j] -s2;
	sp_n_mcc[icolsp][s1][s2]    += (1-delx)*(1-dely);
	sp_n_mcc[icolsp][s1+1][s2]  += delx*(1-dely);
	sp_n_mcc[icolsp][s1][s2+1]  += (1-delx)*dely;
	sp_n_mcc[icolsp][s1+1][s2+1]+= delx*dely;
	
	/* note from djc, should double density at appropriate boundaries */	/* take diagnostics */
	if(theRunWithXFlag) {
          mccrate[2][s1][s2]    += (1-delx)*(1-dely);
          mccrate[2][s1+1][s2]  += delx*(1-dely);
          mccrate[2][s1][s2+1]  += (1-delx)*dely;
          mccrate[2][s1+1][s2+1]+= delx*dely;
	} 
      }
    } 
  }
  
  /**************************************/
  /* He+ + He -> .....                  */
  
  if(icollisional && isp==icolsp) { 
    if(theRunWithXFlag) {
      for (j=0; j<ngr; j++)
        for (k=0; k<ngt; k++)
          mccrate[3][j][k] = 0.0;
    }
    icol_extra += np[icolsp]*col_prob_i;
    N = icol_extra;
    icol_extra -= N;
    
    nnp = np[icolsp];
    for(j=0; j< N; j++) {
      index= nnp*frand();
      nnp--;
      temp = r[icolsp][nnp];
      r[icolsp][nnp] = r[icolsp][index];
      r[icolsp][index] = temp;
      
      temp = th[icolsp][nnp];
      th[icolsp][nnp] = th[icolsp][index];
      th[icolsp][index] = temp;
      
      temp = vr[icolsp][nnp];
      vr[icolsp][nnp] = vr[icolsp][index];
      vr[icolsp][index] = temp;
      
      temp = vth[icolsp][nnp];
      vth[icolsp][nnp] = vth[icolsp][index];
      vth[icolsp][index] = temp;
      
      temp = vz[icolsp][nnp];
      vz[icolsp][nnp] = vz[icolsp][index];
      vz[icolsp][index] = temp;
    }
    
    for(j=nnp; j<nnp+N; j++) {
      maxwellv(&vneutx, &vneuty, &vneutz, vgth);
      vneutx /= vrscale[ionsp];
      vneuty /= vthscale[ionsp];
      vneutz /= vzscale[ionsp];
      
      vr[icolsp][j] -= vneutx;
      vth[icolsp][j] -= vneuty;
      vz[icolsp][j] -= vneutz;
      engy = Exscale[icolsp]*vr[icolsp][j]*vr[icolsp][j];
      engy+= Eyscale[icolsp]*vth[icolsp][j]*vth[icolsp][j];
      engy+= Ezscale[icolsp]*vz[icolsp][j]*vz[icolsp][j];
      vel  = sqrt(2.0*engy/m[icolsp]);
      engy*=6.24219725e18;
      sigma_total = max_sigmav_i/(vel + DBL_MIN);   /**djc**/
      random= frand();
      
      /***************************************/
      /* if the collision is charge exchange */
      
      if (random <= hsigma4(engy)/sigma_total) {
	vr[icolsp][j] = vth[icolsp][j] = vz[icolsp][j] = 0.0;
	
	/* collect diagnostics */
	if(theRunWithXFlag) {
	  s1 = r[icolsp][j]-rminn;
	  s2 = th[icolsp][j];
	  delx = r[icolsp][j]-rminn -s1;
	  dely = th[icolsp][j] -s2;
	  mccrate[3][s1][s2]    += (1-delx)*(1-dely);
          mccrate[3][s1+1][s2]  += delx*(1-dely);
          mccrate[3][s1][s2+1]  += (1-delx)*dely;
          mccrate[3][s1+1][s2+1]+= delx*dely;
        }
      }
      vr[icolsp][j] += vneutx;
      vth[icolsp][j] += vneuty;
      vz[icolsp][j] += vneutz;
    }
  }
}

/**************************************************************/

void hnewvel(energy, vel, vx, vy, vz, e_flag)
     int e_flag;
     SCALAR energy, vel;
     SCALAR *vx, *vy, *vz;
{
  SCALAR phi1, cosphi, sinphi, coschi, sinchi, up1, up2, up3;
  SCALAR mag, r11, r12, r13, r21, r22, r23, r31, r32, r33;

  if(energy < 1e-30)  coschi = 1;
  else  coschi = (energy +2 -2*pow(energy +1,frand()))/energy;
  sinchi= sqrt(fabs(1. - coschi*coschi));
  
  phi1  = 2*M_PI*frand();
  cosphi= cos(phi1);
  sinphi= sin(phi1);
  
  if(e_flag) vel *= sqrt(1.0 - 2.0*m[ecolsp]*(1-coschi)/m[ionsp]);
  
  r13 = *vx;
  r23 = *vy;
  r33 = *vz;
  
  if(r33 == 1.0) { up1= 0;  up2= 1;  up3= 0; }
  else           { up1= 0;  up2= 0;  up3= 1; }
  
  r12 = r23*up3 -r33*up2;
  r22 = r33*up1 -r13*up3;
  r32 = r13*up2 -r23*up1;
  mag = sqrt(r12*r12 + r22*r22 + r32*r32);
  
  r12/= mag;
  r22/= mag;
  r32/= mag;
  
  r11 = r22*r33 -r32*r23;
  r21 = r32*r13 -r12*r33;
  r31 = r12*r23 -r22*r13;
  
  *vx= vel*(r11*sinchi*cosphi +r12*sinchi*sinphi +r13*coschi);
  *vy= vel*(r21*sinchi*cosphi +r22*sinchi*sinphi +r23*coschi); 
  *vz= vel*(r31*sinchi*cosphi +r32*sinchi*sinphi +r33*coschi);
}

/***********************************/
/*  e + He -> e + He  Elastic      */

SCALAR hsigma1(energy)
SCALAR energy;
{
  int i;
  SCALAR engy, els_sigma, alpha;
  static int init_flag=1;
  static SCALAR elastic[NEMAX];

  /******************  Initialization   ********************/
  if(init_flag)
  {
    for(i=0; i<NEMAX; i++)
    {
      engy = i;
      elastic[i] = 8.5e-19/(pow(engy+10.0, 1.1));
    }
    init_flag =0;
  }
  i = energy;
  if(i < NEMAX-1) {
    alpha= energy -i;
    els_sigma = elastic[i] + alpha*(elastic[i+1] - elastic[i]);
  }
  else els_sigma = elastic[NEMAX-1];
  return(els_sigma);
}

/************************************/
/*  e + He -> e + He  Excitation    */

SCALAR hsigma2(energy)
SCALAR energy;
{
  int i;
  SCALAR ext_sigma, engy, alpha;
  static int init_flag=1;
  static SCALAR excit[NEMAX];

  /******************  Initialization   ********************/
  if(init_flag)
  {
    for(i=0; i<NEMAX; i++)
    {
      engy = i;
      if(engy < extengy0 ) excit[i] = 0;
      else if(extengy0 <= engy && energy <27.0)
        excit[i] = 2.08e-22*(engy -extengy0);
      else excit[i] = 3.4e-19/(engy +200);
    }
    init_flag =0;
  }
  i = energy;
  if(i < NEMAX-1) {
    alpha= energy -i;
    ext_sigma = excit[i] + alpha*(excit[i+1] - excit[i]);
  }
  else ext_sigma = excit[NEMAX-1];
  return(ext_sigma);
}

/************************************/
/*  e + He -> e + e + He+  Ion.     */

SCALAR hsigma3(energy)
SCALAR energy;
{
  int i;
  static int init_flag=1;
  static SCALAR ioniz[NEMAX];
  SCALAR ion_sigma, engy, alpha;

  /******************  Initialization   ********************/
  if(init_flag)
  {
    for(i=0; i<NEMAX; i++)
    {
      engy = i;
      if(engy < ionengy0) ioniz[i] = 0;
      else
        ioniz[i] = 1e-17*(engy -ionengy0)/((engy +50)*pow(engy+300.0, 1.2));
    }
    init_flag =0;
  }
  i = energy;
  if(i < NEMAX-1) {
    alpha= energy -i;
    ion_sigma = ioniz[i] + alpha*(ioniz[i+1] - ioniz[i]);
  }
  else
    ion_sigma = ioniz[NEMAX-1];

  return(ion_sigma);
}

/************************************/
/*  He + He+ -> He+ + He  Charge X  */

SCALAR hsigma4(energy)
SCALAR energy;
{
  return(3e-19);
}

int hmakethefile()
{
  SCALAR e;
  FILE *DMPFile;

  DMPFile = fopen("xsections", "w");

  for(e=0; e<300; e+=.1)
    fprintf(DMPFile, "%g %g %g %g %g\n", e+1e-30,
            hsigma1(e)+1e-30, hsigma2(e)+1e-30, hsigma3(e)+1e-30, hsigma4(e)+1e-30); //%g was removed

  fclose(DMPFile);
}

