#include "pdc2.h"
#define NVEL 2000

void load(SCALAR *initn)
{
  int i, j, n, isp, ibin, ibinsrch, nx, ny;
  SCALAR vmag, atheta, aphi, ddv, ccv, rr, fnp;
  SCALAR fcum[NBIN+1], gg[NBIN+1], vv[NBIN+1], vvmag[NVEL];
  SCALAR temp, temp1;
/* attempt to load an annulus */
  SCALAR r0n,r1n;


  r0n = rminn;
  r1n = rmaxn;
  
/* attempt to simulate diocotron */

  /************************************************/
  /* Load one species and one direction at a time */
  for (isp=0; isp<nsp; isp++) {
    np[isp] = initn[isp]*M_PI*(rmax*rmax-rmin*rmin)*zlength/nc2p;
    if (np[isp] > maxnp[isp]) {
      printf("LOAD: too many particles, species %d\n",isp+1);
      exit(1);
    }
    
    /**********************************************************/
    /* quiet load.  vmag loaded uniformly and converted into
       vx, vy, vz using bit reversed random numbers of different
       bases.  x and y are also loaded with  bit reversed random
       numbers of other bases. */
    
    /*****************************************/
    /* calculate the cumulative distribution */
/*-----
    if (vt[isp]>0.0) 
    {
      ibinsrch=3;
      ccv=2.0/(sqrt(TWOPI)*vt[isp]*vt[isp]*vt[isp]);
      ddv=NVTS*vt[isp]/NBIN;
      vv[0]=0.0;
      vv[NBIN]=NVTS*vt[isp];
      gg[0]=0.0;
      gg[NBIN]=0.0;
      for (ibin=1; ibin<NBIN; ibin++) 
      {
        vv[ibin]=ibin*ddv;
        gg[ibin]=ccv*vv[ibin]*vv[ibin]*exp(vv[ibin]*vv[ibin]/
                 (-2.0*vt[isp]*vt[isp]));
      }
      fcum[1]=0.0;
      fcum[0]=0.0;
      fcum[NBIN]=1.0;
      for (ibin=1; ibin<NBIN; ibin++)
        fcum[ibin]=fcum[ibin-1]+(ddv/2.0)*(gg[ibin-1]+gg[ibin]);

      ibinsrch=3;
      for(n=0; n< NVEL; n++) {
	rr=(1.0*n)/NVEL;
	if(ibinsrch>=3) ibinsrch-=3; 
	while (fcum[ibinsrch+1]< rr) ibinsrch += 1;
	vvmag[n]=(vv[ibinsrch]*(fcum[ibinsrch+1]-rr)+
          vv[ibinsrch+1]*(rr-fcum[ibinsrch]))/(fcum[ibinsrch+1]-fcum[ibinsrch]);
      }
    }
------------*/ 
    /*********************************************/
    /* Loading the velocities and positions.   */
    
    for(n=0; n<np[isp]; n++) {
      if (vt[isp]>0.0) {
	vr[isp][n]=v0x[isp] + vt[isp]*random_maxwell();
	vth[isp][n]=v0y[isp] + vt[isp]*random_maxwell();
	vz[isp][n]=v0z[isp] + vt[isp]*random_maxwell();
	
/*-----------
	j = NVEL*revers(n+i,5);
	vmag = vvmag[j];
	atheta=M_PI*revers(n+i, 19);
	aphi=TWOPI*revers(n+i, 7);
	vr[isp][n]+=vmag*sin(atheta)*cos(aphi);
	vth[isp][n]+=vmag*sin(atheta)*sin(aphi);
	vz[isp][n]+=vmag*cos(atheta);
-----------*/
      }
      else { /* rotate thru an arbitrary angle to randomize phases */
	aphi=TWOPI*revers(n+i, 13);
	vr[isp][n] =  cos(aphi)*v0x[isp] + sin(aphi)*v0y[isp];
	vth[isp][n] = -sin(aphi)*v0x[isp] + cos(aphi)*v0y[isp];
	vz[isp][n]=v0z[isp];
      }	
      vr[isp][n] /= vrscale[isp];
      vth[isp][n] /= vthscale[isp];
      vz[isp][n] /= vzscale[isp];
    }
    if (!loader[isp]) 
      for (n=0; n<np[isp]; n++) {
        r[isp][n] = sqrt(r0n*r0n+(n+1.0)*(r1n*r1n-r0n*r0n) /
                    ((SCALAR)np[isp] +1.0));
        th[isp][n] = nct*revers(n+i,3);
        x[isp][n] = r[isp][n] * cos(th[isp][n] *dth);
        y[isp][n] = r[isp][n] * sin(th[isp][n] *dth);
      }
    else 
      for (j=0;j<loader[isp];j++) 
        for (n=np[isp]*j/loader[isp],i=0;n<np[isp]*(j+1)/loader[isp];i++,n++) {
          r[isp][n] = sqrt( sr0[isp][j]*sr0[isp][j] + 
                loader[isp]*i*(sr1[isp][j]*sr1[isp][j]-sr0[isp][j]*sr0[isp][j])/
                ((SCALAR)np[isp]));
          th[isp][n] = sth_width[isp][j]*revers(n+i,3)+sth0[isp][j];
          x[isp][n] = r[isp][n] * cos(th[isp][n] *dth);
          y[isp][n] = r[isp][n] * sin(th[isp][n] *dth);
        }
        
    /***************************************************/
    /* if one particle, stick it in center and give it */
    /* drift + random thermal part                     */
    if (np[isp]==1) {
      x[isp][0]=vt[isp];
      y[isp][0]=0.0;
      r[isp][0] = sqrt(x[isp][0]*x[isp][0]+y[isp][0]*y[isp][0]);
      th[isp][0] = atan2(y[isp][0],x[isp][0]+DBL_MIN);
/*      vr[isp][0]=(v0x[isp]+2.0*(frand()-0.5)*vt[isp])/vrscale[isp];
      vth[isp][0]=(v0y[isp]+2.0*(frand()-0.5)*vt[isp])/vthscale[isp];*/
	temp1= 1.925565e10*dt/2.0;
      vr[isp][0]=sin(temp1)*v0y[isp]/vrscale[isp];
      vth[isp][0]=cos(temp1)*v0y[isp]/vthscale[isp];
      vz[isp][0]=(v0z[isp]+2.0*(frand()-0.5)*vt[isp])/vzscale[isp];
    }
  }
}

/**********************************************************************/

SCALAR bit_rever_maxwell()
{
        static int iset=0;
        static int count=0;
        static SCALAR gset;
        SCALAR fac,r,v1,v2;

        if(iset==0){
          do {
             v1=2.0*revers(count++,2)-1.0;
             v2=2.0*revers(count,5)-1.0;
             r = v1*v1+v2*v2;
          } while (r>=1.0);
          fac=sqrt(-2*log(r)/r);
          gset=v1*fac;
          iset = 1;
          return v2*fac;
        }
        else{
          iset=0;
          return gset;
        }
}

SCALAR random_maxwell()
{
        static int iset=0;
        static SCALAR gset;
        SCALAR fac,r2,v1,v2;

        if(iset==0){
                do{
                        v1=2.0*frand()-1.0;
                        v2=2.0*frand()-1.0;
                        r2 = v1*v1+v2*v2;
                } while (r2>=1.0);
                fac=sqrt(-2*log(r2)/r2);
                gset=v1*fac;
                iset = 1;
                return v2*fac;
        }
        else{
                iset=0;
                return gset;
        }
}

/**********************************************************************/

