#include "pdc2.h"
#define NVEL 2000

/****************************************************************/
/* Routine to adjust (initialize, re-pack and inject) particles */
/*  to the desired boundary conditions                          */
/* Not yet 1/26/93, PM                                          */
/* added checks if particle does bad stuff with polar coords.   */
/* 2/2/93 changed it to be more like pdp2-periodic. fixed a bug */
/* or three. Specifically, added x,y,vx,vy when changing stuff  */
/* Added Injection of particles	P.G.				*/
/****************************************************************/
int adjust(isp)
int isp;
{
  register int n,k,ii,nnp,lost,secountl, secountr;
  static SCALAR extra[NSMAX];
  static int init_flag=1;
  SCALAR del_t;
  int nreflux[NSMAX], ninjected,ntheta;
  extern double drand48();
  static int seed = 0, load=0;
  int first = 1, count = 0;
  SCALAR aphi;
  SCALAR tmp = 0.0;
  int j,ibinsrch, ibin;

  /* INITIALIZE array for computing positions of injected particles */
  if (init_flag)
  {
    /* "enter" is now the no. of particles injected each "dt" */
    for (ii=0; ii<nsp; ii++)
    {
      if (fabs(enter[ii]) > 0.) extra[ii] = 0.5123123;
      else extra[ii] = 0.;
    }
    init_flag = 0;
  }
  
  jwall[isp] = 0.0;
  if(secondary) secountl = secountr = 0;

    nreflux[isp]= 0;    /* nreflux[isp] equals zero */
    if(np[isp]>0) {
      nnp = np[isp]-1;
      n=0;
      do {

       if ( r[isp][n] <= rminn && cent_cond) {
         if (BCflag==0) { 
          /* lost particle to inner wall, so get rid of it cleverly
            and we don't need to calculate r and th */
          /*PG modified */
           r[isp][n] =  r[isp][nnp];
           th[isp][n] = th[isp][nnp];
           vr[isp][n] = vr[isp][nnp];
           vth[isp][n] = vth[isp][nnp];
           vz[isp][n] = vz[isp][nnp];
           if (reflux)  nreflux[isp]++;
           nnp--;
           jwall[isp] += q[isp]/(TWOPI*rmin*zlength);  /* Wall Charge Density */
           if (secondary)
             if (frand() < seec[isp]) {
               secountl++;
               jwall[secsp] -= q[secsp]/(TWOPI*rmin*zlength);
             }
         }
         else {        /* for the case of Neumann B.C. */
           r[isp][n] =1.01*rminn;
           tmp = (vr[isp][n]) ? atan(vth[isp][n]/vr[isp][n])/PI : 1;
           th[isp][n] += (tmp-0.5)*nct;
           while(th[isp][n]>=nct) th[isp][n]-=nct;
           vr[isp][n] = - vr[isp][n];
           vth[isp][n] =  vth[isp][n];
         }
        }
       else if (r[isp][n] >= rmaxn) {
	  /* 
	    lost particle to wall, so get rid of it cleverly
	    and we don't need to calculate r and th 
	    */
	  /*djc modified */
	   r[isp][n] =  r[isp][nnp];
	   th[isp][n] =  th[isp][nnp];
	   vr[isp][n] = vr[isp][nnp];
	   vth[isp][n] = vth[isp][nnp];
           vz[isp][n] = vz[isp][nnp];
           nnp--;
           if (reflux) nreflux[isp]++;	        /* leehj */
          if (secondary) if (frand() < seec[isp])    secountr++;
	}
	else n++;
      } while(n <= nnp);
      np[isp] = nnp+1;
    }


  /* INJECT new particles at walls, one species at a time */
  
    if (np[isp]>=inject_max_np[isp]) inject[isp]=0;
    else if (inject[isp])
    {
      if (loader[isp]==0)      /* source region  , 鿡 Ի */
      {
        extra[isp] += enter[isp]*sp_k[isp];
        ninjected = extra[isp];

        while (extra[isp] >= 1.)
        {
          del_t= ((int)extra[isp])/(ninjected +1.);
          ii = np[isp];
          np[isp]++;

          if (ii > maxnp[isp]) /* Move array boundaries here */
          {
            printf("ADJUST: too many particles, species %d",isp);
            exit(1);
          }
      
	  ntheta =  th[isp][ii] = nct*revers(seed,2);
          vr[isp][ii] = v0x[isp] + vt[isp]*random_maxwell();
          vth[isp][ii] = v0y[isp] + vt[isp]*random_maxwell();
          vz[isp][ii] = v0z[isp] + vt[isp]*random_maxwell();
          vr[isp][ii]+= qm[isp]*(del_t-0.5)*er[0][ntheta]*dt;
          vth[isp][ii]+= qm[isp]*(del_t-0.5)*eth[0][ntheta]*dt;
          vr[isp][ii] /= vrscale[isp];
          vth[isp][ii] /= vthscale[isp];
          vz[isp][ii] /= vzscale[isp];
          r[isp][ii] = rminn + del_t*vr[isp][ii];
	  if (r[isp][ii]<0) r[isp][ii]=0;
          x[isp][ii] = r[isp][ii]*cos(th[isp][ii]*dth);
          y[isp][ii] = r[isp][ii]*sin(th[isp][ii]*dth);
          jwall[isp] -= q[isp]/(TWOPI*rmin*zlength);
          extra[isp] -= 1;
          seed++; 
        }       /* end "while extra..)" loop   */
      }         /* end "if (loader[isp]==0)" loop */
      
      /*---------------------------------------------------------------------*/
      else {    /* for the case of source region is defined */
        extra[isp] += enter[isp]*sp_k[isp];
        ninjected = extra[isp];
        while (extra[isp] >= 1.) {
          ii = np[isp]; 
          np[isp]++;
          if (ii > maxnp[isp]) {
            printf("ADJUST: too many particles, species %d",isp);
            exit(1);
          }

  	  th[isp][ii] = sth0[isp][load]+sth_width[isp][load]*frand();
          r[isp][ii] = sr0[isp][load] + r_width*frand() +
               (sr1[isp][load]-sr0[isp][load])*th[isp][ii]/sth_width[isp][load];
          if (frand()>r[isp][ii]/(sr1[isp][load]+r_width)) {
	   np[isp]--;
           continue;
          } 
          while(th[isp][ii]>ngt) th[isp][ii]-=ngt;
          x[isp][ii] = r[isp][ii]*cos(th[isp][ii]*dth);
          y[isp][ii] = r[isp][ii]*sin(th[isp][ii]*dth);

          if (vt[isp]>0.0) {
            vr[isp][ii]=v0x[isp]+vt[isp]*random_maxwell();
            vth[isp][ii]=v0y[isp]+vt[isp]*random_maxwell();
            vz[isp][ii]=v0z[isp]+vt[isp]*random_maxwell();
          }
          else {
            aphi=TWOPI*frand();
            vr[isp][ii] =  cos(aphi)*v0x[isp] + sin(aphi)*v0y[isp];
            vth[isp][ii] = -sin(aphi)*v0x[isp] + cos(aphi)*v0y[isp];
            vz[isp][ii] =  v0z[isp];
          }
          vr[isp][ii] /= vrscale[isp];
          vth[isp][ii] /= vthscale[isp];
          vz[isp][ii] /= vzscale[isp];
          if (++load==loader[isp]) load=0;
          extra[isp] -= 1;
        }       /* end "while extra..)" loop   */
      } /* end "else" loop          */

    }   /* end "if( inject..)" loop */

    /*-------------------------------------------------*/
    /*                reflux form                      */
    if (nreflux[isp]) while (nreflux[isp]>0) {   
      ii=np[isp]++;
      nreflux[isp]--;
      if (vt[isp]>0.0) {
        vr[isp][ii]=v0x[isp]+vt[isp]*random_maxwell();
        vth[isp][ii]=v0y[isp]+vt[isp]*random_maxwell();
        vz[isp][ii]=v0z[isp]+vt[isp]*random_maxwell();
      }
      else { /* rotate thru an arbitrary angle to randomize phases */
        aphi=TWOPI*frand();
        vr[isp][ii] =  cos(aphi)*v0x[isp] + sin(aphi)*v0y[isp];
        vth[isp][ii] = -sin(aphi)*v0x[isp] + cos(aphi)*v0y[isp];
        vz[isp][ii] =  v0z[isp];
      }
      vr[isp][ii] /= vrscale[isp];
      vth[isp][ii] /= vthscale[isp];
      vz[isp][ii] /= vzscale[isp];
      r[isp][ii] = ncr*frand()+rminn;
      th[isp][ii] = nct*frand();
      x[isp][ii] = r[isp][ii]*cos(th[isp][ii]*dth);
      y[isp][ii] = r[isp][ii]*sin(th[isp][ii]*dth);
    } 

    jwall[isp] /= sp_k[isp]*dt;
 
} /* end ADJUST */


