#define GLOBALORIGIN
#include "pdc2.h"
#undef GLOBALORIGIN

#include <xgrafix.h>

FILE *InputDeck, *DropFile;
char dumpfile[20], dropfile[20];


int display_title();
int start();
int mccdiag_init();
int InitWindows(int,char **);
int history();
int history_nox();
int adjust(int);
int species(int, SCALAR *, SCALAR *);
int Restore(char *);
void load(SCALAR *);
int Dump(char *);

int main(argc, argv)
int argc;
char *argv[];
{
  register int j, k, i;

  display_title();

  /*********************************************************/
  /* Read input file, and initialize params and vars. This */
  /* function calls SPECIES and LOAD or Restore            */
  XGInit(argc,argv,&t);
  start(); 

  if(theRunWithXFlag) {
    mccdiag_init();
    InitWindows(argc, argv);
    iter=1;
  }
/*  InitVars ();		*/

  oldsigma = sigma = 0.0;

  for (i=0;i<nsp;i++) {
    gather (i); 
    for (j=0; j<ngr; j++) 
      for (k=0; k<ngt; k++) srho[i][j][k]=srho1[i][j][k]; 
  }
  fields();
  exti=0;
  if(theRunWithXFlag) history();
  else		      history_nox();
/*
  strcpy(dumpfile, argv[1]);  strcat(dumpfile, ".dmp");
  strcpy(dropfile, argv[1]);  strcat(dropfile, ".drp");
*/
  for(i=0; i<nsp; i++) printf("np[%d] = %d, ", i, np[i]);
  printf("\nvrscale=dr/dt=%e,v_th(mid_r)=%e\n",vrscale[0],(rmax-rmin)*dth/2/dt);
  XGStart();

  return 0;
}

void XGMainLoop()
{  
   register int i, j, isp;
   SCALAR frac, temp;
   t += dt; iter++;
   
   for (isp=0;isp<nsp;isp++) {   
     if(!(k_count[isp]%sp_k[isp])) {
       (*move_ptr)(isp);
       adjust(isp);
       if (ecollisional || icollisional) (*mccptr)(isp);
       gather (isp);
       k_count[isp]=0;
     }
     k_count[isp]++;
    frac = ((SCALAR)k_count[isp])/sp_k[isp];
    for (i=0;i<ngr;i++) 
      for (j=0;j<ngt;j++) {
        srho[isp][i][j] = (1-frac)*srho0[isp][i][j] + frac*srho1[isp][i][j]
                          +sp_n_mcc[isp][i][j];
      }
   }
   fields();
   if(theRunWithXFlag) history();
   else 	       history_nox();
/*
   temp=0;
   for (i=0;i<np[0];i++) temp+=r[0][i];
   printf("CEM=%g\n",temp*dr/np[0]);
*/    
}

/***********************************************************/

int start()
{
  register int i, j, k,isp;
  SCALAR initn[NSMAX], rmax_ion, tmp1, tmp2, tmp3;
  int dust_mode;
  char aachar[80];
  SCALAR jj0[NSMAX], s_area; /* only injection from left currently */

  /***********************************************/
  /* Open files and read in "general" parameters */

  if (!WasInputFileGiven) {
    puts("No input file!!");
    exit(-1);
  }
  else InputDeck = fopen(theInputFile, "r");


/* default no center conductor */

   cent_cond = 0;

  /* Read lines until we get to numbers */
#ifndef XG_SCALAR_DOUBLE
   while (fscanf(InputDeck,"%d %d %d %g %g %g %g", 
#else
       while (fscanf(InputDeck,"%d %d %d %lg %lg %lg %lg", 
#endif
	&nsp, &ncr, &nct, &nc2p, &dt, &zlength, &epsilon) <7)
    fscanf(InputDeck, "%s", aachar);

#ifndef XG_SCALAR_DOUBLE
while (fscanf(InputDeck,"%g %g %d %g %g %g", 
#else
    while (fscanf(InputDeck,"%lg %lg %d %lg %lg %lg", 
#endif
  &inner_dc, &inner_ac, &inner_mode, &inner_f0, &delta1, &inner_t0) <6)
    fscanf(InputDeck, "%s", aachar);

#ifndef XG_SCALAR_DOUBLE
while (fscanf(InputDeck,"%g %g %d %g %g %g", 
#else
    while (fscanf(InputDeck,"%lg %lg %d %lg %lg %lg", 
#endif
  &outer_dc, &outer_ac, &outer_mode, &outer_f0, &delta2, &outer_t0) <6)
    fscanf(InputDeck, "%s", aachar);

  inner_ac*=0.5;
  outer_ac*=0.5;

#ifndef XG_SCALAR_DOUBLE
  while (fscanf(InputDeck, "%g %g %g %g %g %g %g",
#else
      while (fscanf(InputDeck, "%lg %lg %lg %lg %lg %lg %lg",
#endif
/*	&rhoback,&bmag,&bphi,&btheta,&rmin,&rmax,&dust_mode) <7)	*/
	&rhoback,&bmag,&bphi,&btheta,&rmin,&rmax,&rmax_ion) <7)
    fscanf(InputDeck, "%s", aachar);

#ifndef XG_SCALAR_DOUBLE
while (fscanf(InputDeck, "%d %d %g %d %d %d",
#else
    while (fscanf(InputDeck, "%d %d %lg %d %d %d",
#endif
	&modsrt,&modend,&moder, &nfft, &cflag, &BCflag) <6)
    fscanf(InputDeck, "%s", aachar);
  
/* Circuitry */

  while (fscanf(InputDeck, "%lf %lf %lf %lf %lf %lf %d",
        &backj, &extr, &extl, &extc, &extq, &bias_volt, &bias_grid) < 7)
    fscanf(InputDeck, "%s", aachar);

/* Collisonal Params */

  while (fscanf(InputDeck, "%d %d %d %d %d %d",
        &gas, &secsp, &ecollisional, &icollisional, &reflux, &nsmooth) <6)
    fscanf(InputDeck, "%s", aachar);

  if (gas==1) printf("gas %d : Helium\n",gas);
  if (gas==2) printf("gas %d : Argon\n",gas);
  
#ifndef XG_SCALAR_DOUBLE
  while (fscanf(InputDeck, "%g %g %d %g %g", 
#else
      while (fscanf(InputDeck, "%lg %lg %d %lg %lg", 
#endif
      &seec[0], &seec[1], &ionsp, &pressure, &gtemp) <5)
    fscanf(InputDeck, "%s", aachar);

  /* Check for errors in the "general" input parameters */

  if (extl < 0. || extr < 0. || extc < 0.)
  {
    printf("START: dont be cute with R,L,C < 0\n");
    exit(1);
  }

  if (ncr<2 || nct<2)
  {
    puts("START: dont be cute with ncr,nct<2\n");
    exit(1);
  }

  if (dt <= 0.0) {
    puts("START: dt <= 0!!\n");
    exit(1);
  }
   
  if(modsrt == 0 || modend >= ncr/2){
    puts("START: moden value not valid\n");
    exit(1);
  }
  mccptr =argonmcc;
  switch (gas) {
    case HELIUM:
         mccptr =heliummcc;
         break;
    case ARGON:
         mccptr =argonmcc;
         break;
  }
  move_ptr=move;
  if (bmag>0.0) move_ptr = mag_move;

  ngr = ncr+1;  
  ngt = nct+1; 
  ngfactor_r=2;
  ngfactor_t=4;
  hngr=ncr/ngfactor_r+1;
  hngt=nct/ngfactor_t+1;
  fncr= (SCALAR)ncr; 
  fnct= (SCALAR)nct;
  fftngt = nct/2;

  num_modes = modend-modsrt+1;
/*  num_modes2=num_modes*2; */

  dt2 = dt/2;
  dr = (rmax-rmin)/fncr; 
  dth = TWOPI/fnct; 
  dr2 = dr/2;
  dth2 = dth/2;
  if (moder>rmax || moder < rmin) {
     puts ("modeR should be less than rmax !!");
     exit(-1);
  }
  else moder=(moder-rmin)/dr;
/* djc def to avoid arctan somewhat arbitrary */
  max_ang= dth*dth*dth;

#ifdef DEBUG1
  printf("ngr = %d, ngt = %d, fftngt = %d\n",ngr,ngt,fftngt);
  printf("dr = %f, dth = %f\n",dr,dth);
#endif
  interval=1;

  /***************************************/
  /* Allocate space for field parameters */
  
  rg = (SCALAR  *)malloc(ngr*sizeof(SCALAR));
  r_array= (SCALAR *)malloc(ngr*sizeof(SCALAR));
  r_array2= (SCALAR *)malloc(hngr*sizeof(SCALAR));
  for(i=0; i< ngr; i++) r_array[i] = rmin + i*dr;
  for(i=0; i< hngr; i++) r_array2[i]=r_array[i*ngfactor_r];
  
  th_array= (SCALAR *)malloc(ngt*sizeof(SCALAR));
  th_array2= (SCALAR *)malloc(hngt*sizeof(SCALAR));
  cos_array= (SCALAR *)malloc(ngt*sizeof(SCALAR));
  sin_array= (SCALAR *)malloc(ngt*sizeof(SCALAR));
  for(i=0; i< ngt; i++) {
    th_array[i] = i*dth;
    cos_array[i] = cos(i*dth);
    sin_array[i] = sin(i*dth);
    }
  for(i=0; i< hngt; i++)  th_array2[i]=th_array[ngfactor_t*i];
  k_array= (SCALAR *)malloc(fftngt*sizeof(SCALAR));
  for(i=0; i< fftngt; i++) 
    k_array[i] = i;

/* draw the inner cylinder */
  if (rmin > 0.0)
  {
  cent_cond = 1;
  boundn=150;
  boundxmn= (SCALAR *)malloc(boundn*sizeof(SCALAR));
  boundymn= (SCALAR *)malloc(boundn*sizeof(SCALAR));
    for(i=0; i< boundn; i++) {
      boundxmn[i] = rmin/dr*cos(i*2*M_PI/boundn);
      boundymn[i] = rmin/dr*sin(i*2*M_PI/boundn);
    }
  }

/* the outer cylinder */

  boundn=150;
  boundxmx= (SCALAR *)malloc(boundn*sizeof(SCALAR));
  boundymx= (SCALAR *)malloc(boundn*sizeof(SCALAR));
  for(i=0; i< boundn; i++) {
    boundxmx[i] = rmax/dr*cos(i*2*M_PI/boundn);
    boundymx[i] = rmax/dr*sin(i*2*M_PI/boundn);
  }
  dust_rho= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
/*-------------
  rhoback/=(rmax-rmin)*(rmax-rmin)/6;
  for (i=0; i< ngr; i++) {
    dust_rho[i]=(SCALAR *)malloc(ngt*sizeof(SCALAR));
    for (j=0; j<ngt; j++) {
        dust_rho[i][j]=rhoback*(rmax-r_array[i])*(r_array[i]-rmin);
      if (dust_mode) 
        dust_rho[i][j]*=2*cos(dust_mode*th_array[j]/2)
                         *cos(dust_mode*th_array[j]/2);
    }
  }
---------------*/
  for (i=0; i< ngr; i++) {
    dust_rho[i]=(SCALAR *)malloc(ngt*sizeof(SCALAR));
    for (j=0; j<ngt; j++) {
	if (r_array[i]<rmax-rmax_ion) dust_rho[i][j]=rhoback;
	else dust_rho[i][j]=0;
    }
  }

  srho= (SCALAR ***)malloc(nsp*sizeof(SCALAR **));
  srho0= (SCALAR ***)malloc(nsp*sizeof(SCALAR **));
  srho1= (SCALAR ***)malloc(nsp*sizeof(SCALAR **));
  sp_n_mcc= (SCALAR ***)malloc(nsp*sizeof(SCALAR **));

  for(isp=0; isp<nsp; isp++)
  {
    srho[isp]= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
    srho0[isp]= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
    srho1[isp]= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
    sp_n_mcc[isp]= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
    for(i=0; i< ngr; i++) {
      srho[isp][i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));
      srho0[isp][i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));
      srho1[isp][i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));
      sp_n_mcc[isp][i]= (SCALAR *)malloc(ngt*sizeof(SCALAR *));
      for (j=0;j<ngt; j++) srho0[isp][i][j]=
                         srho1[isp][i][j]=sp_n_mcc[isp][i][j]=srho[isp][i][j]=0;
    }
  }
/*------------------ 2D Diagnostics ----------------------------*/
  phi_r_show=(SCALAR *)malloc(ngr*sizeof(SCALAR));
  er_r_show=(SCALAR *)malloc(ngr*sizeof(SCALAR));
  eth_r_show=(SCALAR *)malloc(ngr*sizeof(SCALAR));
  srho_r_show=(SCALAR **)malloc(nsp*sizeof(SCALAR *));
  for (isp=0;isp<nsp;isp++) 
    srho_r_show[isp]=(SCALAR *)malloc(ngr*sizeof(SCALAR));
  sr0=(SCALAR **)malloc(nsp*sizeof(SCALAR *));
  sth0=(SCALAR **)malloc(nsp*sizeof(SCALAR *));
  sr1=(SCALAR **)malloc(nsp*sizeof(SCALAR *));
  sth_width=(SCALAR **)malloc(nsp*sizeof(SCALAR *));
/*--------------------------------------------------------------*/

/* _i for irregular mesh */

  x_i= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    x_i[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  y_i= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    y_i[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  /* all the rho's */
  rho= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    rho[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  rhok= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    rhok[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  fft_rho= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    fft_rho[i] = (SCALAR *)malloc(fftngt*sizeof(SCALAR));

  /* And then the phi's */
  phi= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    phi[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  phik= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    phik[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  fft_phi= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    fft_phi[i] = (SCALAR *)malloc(fftngt*sizeof(SCALAR));

/* _i for irregular mesh */

  phi_i= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    phi_i[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  er= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    er[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  eth= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    eth[i]= (SCALAR *)malloc(ngt*sizeof(SCALAR));

  /*------------leehj--------------------------*/
  er2= (SCALAR **)malloc(hngt*sizeof(SCALAR *));
  for(i=0; i< hngt; i++) er2[i] = (SCALAR *)malloc(hngr*sizeof(SCALAR));

  eth2= (SCALAR **)malloc(hngt*sizeof(SCALAR *));
  for(i=0; i< hngt; i++) eth2[i]= (SCALAR *)malloc(hngr*sizeof(SCALAR));
  /*-------------------------------------------*/

  ar= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    ar[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  ath= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++) 
    ath[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  sigma_wall= (SCALAR *)malloc(ngt*sizeof(SCALAR));

  gam  = (SCALAR *)malloc(ngr*sizeof(SCALAR));
  
 /* Coefficients for tridiagonal field solve, PM */ 

  A = (SCALAR *)malloc(ngr*sizeof(SCALAR));
  C = (SCALAR *)malloc(ngr*sizeof(SCALAR));
  for (j=1; j< ngr; j++)
  { 
    A[j] = -1.0/(2.0*r_array[j]*dr)+1.0/(dr*dr);
    C[j] =  1.0/(2.0*r_array[j]*dr)+1.0/(dr*dr);
  }
  if (bias_grid>0 && bias_grid<ngr) A[bias_grid]=C[bias_grid]=0;
  C[0] = -1;
  
  B = (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0;i < ngr;i++) 
    B[i]=(SCALAR *)malloc(ngt*sizeof(SCALAR));

  for(k=0;k < ngt; k++)
    B[0][k] = 1;

  for(j=1;j< ngr;j++)
    for(k=0; k< fftngt; k++)
      B[j][2*k]  = 
      B[j][2*k+1]= -2.0/(dr*dr)-(4.0*sin(M_PI*k/nct)*sin(M_PI*k/nct))/
                                (r_array[j]*r_array[j]*dth*dth);
  if(bias_grid>0 && bias_grid<ngr) 
    for(k=0; k< fftngt; k++) B[bias_grid][2*k]=B[bias_grid][2*k+1]=1;

  for(i=0; i < ngr;i++)
    for(j=0; j < ngt;j++) {
      x_i[i][j] = r_array[i]*cos(th_array[j]);
      y_i[i][j] = r_array[i]*sin(th_array[j]);
    }

  /* Allocate space for time history diagnostics */

  t_array = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  dt_hist = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  cem_hist = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  a_momentum_hist = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));

  k_count = (int *)malloc(nsp*sizeof(int *));
  for (i=0;i<nsp;i++) k_count[i]=0;
  sp_k    = (int *)malloc(nsp*sizeof(int *));
  np_hist = (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  for(i=0;i<nsp;i++)
    np_hist[i]=(SCALAR *)malloc(HISTMAX*sizeof(SCALAR));

  ke_hist = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  te_hist = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  kes_hist= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  for(i=0;i<nsp;i++)
    kes_hist[i]=(SCALAR *)malloc(HISTMAX*sizeof(SCALAR));

  fe_hist = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  fe_hist_old = (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));

  fes_hist = (SCALAR **)malloc(num_modes*sizeof(SCALAR *));
  for(i=0;i<num_modes;i++)
    fes_hist[i]=(SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  
  loc_ese = (SCALAR **)malloc(3*sizeof(SCALAR *));
  for(i=0;i<3; i++) loc_ese[i] = (SCALAR *)malloc(fftngt*sizeof(SCALAR));
  
  com_cur_hist= (SCALAR  *)malloc(HISTMAX*sizeof(SCALAR));
  com_phi_hist= (SCALAR **)malloc(3*sizeof(SCALAR *));
  for (i=0;i<3;i++) com_phi_hist[i]= (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));

/* mid Electric Fields */
  com_e_hist= (SCALAR **)malloc(2*sizeof(SCALAR *));
  com_e_hist[0]= (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  com_e_hist[1]= (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));

  sigma_hist= (SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  for(j=0;j<ngt;j++) sigma_wall[j] = 0.0;
/* 
  mode_hist = (SCALAR **)malloc(num_modes2*sizeof(SCALAR *));
  for(j=0;j<num_modes2;j++) 
    mode_hist[j]=(SCALAR *)malloc(HISTMAX/16*sizeof(SCALAR));
*/
  mode_hist=(SCALAR *)malloc(HISTMAX*sizeof(SCALAR));
  if(nfft)
  {
/* attempt to  make spectral values work */
/*   if (cent_cond) */
   {
    freq_hi = nfft/2;
    df = 1/(nfft*dt);
    f_array= (SCALAR *)malloc(nfft*sizeof(SCALAR));
    for(i=0; i< nfft; i++) f_array[i]= i*df;
   }
/*   else
   {
    freq_hi = nfft;
    df = 1/(nfft*dt);
    f_array= (SCALAR *)malloc(nfft*sizeof(SCALAR));
    for(i=0; i< nfft; i++) f_array[i]= i*df -0.5/dt;
   }
*/
    cur_hist= (SCALAR  *)malloc(nfft*sizeof(SCALAR));
    phi_hist= (SCALAR  **)malloc(3*sizeof(SCALAR  *));
    for (i=0;i<3;i++) phi_hist[i]= (SCALAR  *)malloc(nfft*sizeof(SCALAR));

/* local E */
    e_hist= (SCALAR  **)malloc(2*sizeof(SCALAR  *));
    e_hist[0]= (SCALAR  *)malloc(nfft*sizeof(SCALAR));
    e_hist[1]= (SCALAR  *)malloc(nfft*sizeof(SCALAR));


    Local_t_array = (SCALAR *) malloc(nfft*sizeof(SCALAR));

    phi_r_klo_f = (SCALAR *)malloc(nfft*sizeof(SCALAR));
    phi_r_khi_f = (SCALAR *)malloc(nfft*sizeof(SCALAR));
    phi_r_khi2_f = (SCALAR *)malloc(nfft*sizeof(SCALAR));
    phi_r_khi3_f = (SCALAR *)malloc(nfft*sizeof(SCALAR));

    phi_r_klo_t = (SCALAR *)malloc(nfft*sizeof(SCALAR));
    phi_r_khi_t = (SCALAR *)malloc(nfft*sizeof(SCALAR));
/*
    phi_r_kth_t= (SCALAR **)malloc(num_kth*sizeof(SCALAR *));

    for(i=0; i<num_kth; i++)
    {
      phi_r_kth_t[i]= (SCALAR *)malloc(2*nfft*sizeof(SCALAR));
      for(j =0; j<2*nfft; j++)
      {
        phi_r_kth_t[i][j]= 0.0;
      }
    }
*/
/* Time average Diagnostics */

/* Potential */

  phi_ave= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    phi_ave[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  phi_ave_show= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    phi_ave_show[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  for(i=0; i< ngr; i++)
   for(j=0; j< ngt; j++)
     phi_ave[i][j] = phi_ave_show[i][j] = 0.0;

/* Electric Field */

/* Er */
  er_ave= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    er_ave[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  er_ave_show= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    er_ave_show[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  for(i=0; i< ngr; i++)
   for(j=0; j< ngt; j++)
     er_ave[i][j] = er_ave_show[i][j] = 0.0;

/* Eth */
  eth_ave= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    eth_ave[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  eth_ave_show= (SCALAR **)malloc(ngr*sizeof(SCALAR *));
  for(i=0; i< ngr; i++)
    eth_ave_show[i] = (SCALAR *)malloc(ngt*sizeof(SCALAR));

  for(i=0; i< ngr; i++)
   for(j=0; j< ngt; j++)
     eth_ave[i][j] = eth_ave_show[i][j] = 0.0;

/* spectral components */

    cur_fft= (SCALAR  *)malloc(nfft*sizeof(SCALAR));
    phi_fft= (SCALAR  *)malloc(nfft*sizeof(SCALAR));
    er_fft= (SCALAR  *)malloc(nfft*sizeof(SCALAR));
    eth_fft= (SCALAR  *)malloc(nfft*sizeof(SCALAR));

  }
/**************************************/
  

  vol = (SCALAR *)malloc(ngr*sizeof(SCALAR));
  vol2 = (SCALAR *)malloc(ngr*sizeof(SCALAR));
  pidrdr = M_PI*dr*dr;
  drdthzlength = dr*dth*zlength;
  pidrdrzlength = M_PI*dr*dr*zlength;
/*
  if (!cent_cond) vol[0]=pidrdrzlength/3.0;
  else            vol[0]=drdthzlength*(2*r_array[0]+r_array[1])/6.0;
  for (i=1;i<ncr;i++) vol[i]=r_array[i]*drdthzlength;
  vol[ncr]=drdthzlength*(2*r_array[ncr]+r_array[ncr-1])/6.0;
*/
  if (!cent_cond) { 
    vol[0]=M_PI*dr/dth/3.0; 
    vol2[0]=dr/6.0; 
  }
  else { 
    vol[0]=(2*r_array[0]+r_array[1])/6.0;
    vol2[0]=0.5*(r_array[0]+dr/3.0);
  }
  for (i=1;i<ncr;i++) vol[i]=vol2[i]=r_array[i];
  vol[ncr]=(2*r_array[ncr]+r_array[ncr-1])/6.0;
  vol2[ncr]=0.5*(r_array[ncr]-dr/3.0);

  area0  = 0.5; /* area of zeroth cell in normalized coords*/
  area = TWOPI*rmin*zlength;
  delta1 *= TWOPI/360.0;
  delta2 *= TWOPI/360.0;
  epsilon *= EPS0;
  inner_f0 *= TWOPI;
  outer_f0 *= TWOPI;
  epsdth2 = 2*epsilon*dth;
  se = -1/epsilon; /* note the minus sign, used in fields */
  extq_1 = extq_2 = extq_3 = extq;

  /********************************************/
  secondary = secsp;
  secsp--;

  /********************************************/
  /* Set up the parameters for each "species" */

  for (isp=0; isp<nsp; ++isp) {
    species(isp, initn, jj0);
    qm[isp] = q[isp]/m[isp];

    tmp1=sqrt(initn[isp]*q[isp]*qm[isp]/8.85*1e12);
    tmp2=fabs(qm[isp])*bmag/(2*M_PI);
    tmp3=(tmp1) ? vt[isp]/tmp1 : 0;
    printf("Species %d: Initial fp=%g, fc=%g, Debye length = %g\n\n",
	isp+1, tmp1/(2*M_PI), tmp2, tmp3);

/* djc changed 1/4 to 1/2 */
    if(bmag > 0.0)
      a_scale[isp] = 0.5*qm[isp]*sp_k[isp]*sp_k[isp]*dt*dt/dr;
    else
      a_scale[isp] = qm[isp]*sp_k[isp]*sp_k[isp]*dt*dt/dr;
    dt_min3=-0.35/(qm[0]*bmag);
    q[isp] *= nc2p;
    vrscale[isp]= dr/(sp_k[isp]*dt);
    vthscale[isp]= dr/(sp_k[isp]*dt);
    vzscale[isp]= zlength/(sp_k[isp]*dt);

    Exscale[isp] = 0.5*nc2p*vrscale[isp]*vrscale[isp]*m[isp];
    Eyscale[isp] = 0.5*nc2p*vthscale[isp]*vthscale[isp]*m[isp];
    Ezscale[isp] = 0.5*nc2p*vzscale[isp]*vzscale[isp]*m[isp];

/* injected stuff */
    if(enter[isp] > 0.) inject[isp] = 1;     
    if(bmag) {
/*    ttmag = tan(dt2*sp_k[isp]*qm[isp]*bmag);  --------- Original
      ttmag = tan(theta/2) in Boris method, and this notation is 
      valid only if theta is small. Here, theta is rotation angle. -------*/
      ttmag = dt2*sp_k[isp]*qm[isp]*bmag;       /*  LEEHJ  */
      ttx[isp]=ttmag*sin(btheta)*cos(bphi);
      tty[isp]=ttmag*sin(btheta)*sin(bphi);
      ttz[isp]=ttmag*cos(btheta);              
      ssx[isp]=2.0*ttx[isp]/(1.0+ttmag*ttmag);
      ssy[isp]=2.0*tty[isp]/(1.0+ttmag*ttmag); 
      ttx[isp]=tty[isp]=ssx[isp]=ssy[isp]=0;
      ssz[isp]=2.0*ttz[isp]/(1.0+ttmag*ttmag);
    }  
  }   /* end  "for isp=0 thru nsp-1 "  loop */

  r = (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  th = (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  x = (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  y = (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  vr= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  vth= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  vz= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  vx1= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  vy1= (SCALAR **)malloc(nsp*sizeof(SCALAR *));

  w1= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  w2= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  w3= (SCALAR **)malloc(nsp*sizeof(SCALAR *));
  w4= (SCALAR **)malloc(nsp*sizeof(SCALAR *));

  for(isp=0; isp<nsp; isp++)
  {
    r[isp] = (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    th[isp] = (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    x[isp] = (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    y[isp] = (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    vr[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    vth[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    vz[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    vx1[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    vy1[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));

    w1[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    w2[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    w3[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
    w4[isp]= (SCALAR *)malloc(maxnp[isp]*sizeof(SCALAR));
  }

  if(nsp && (!r[nsp-1] || !th[nsp-1] || !vr[nsp-1] || !vth[nsp-1] || !vz[nsp-1]))
  {
    puts("START: x, y, v malloc failed");  exit(1);
  }


  
/* test to see if Debye length criterion is met, P.G. */
/* #ifdef DEBYE */
  for (isp=0; isp<nsp; ++isp) {
	T[isp] = fabs (0.5*vt[isp]*vt[isp]/qm[isp]); /* eV */
	LambdaD[isp] = 7.43e+3*sqrt(T[isp]/initn[isp]); /* N in m-3 */ 

	if (T[isp] && dr > 2.0* max(LambdaD[0],LambdaD[1])) 
	{
         puts("Warning!, Grid Spacing > Debye length\n");
	 printf ("Recommended ncr = %d\n",
		(int) ((rmax-rmin)/max(LambdaD[0],LambdaD[1])) );
	}
  }
/* #endif  */



  /********************************************************/
  /* Load in all species' particles, properly distributed */
  /* Start from a dump file if given otherwise load the   */
  /* system with the initial distribution                 */ 
 
  /* normalize rmax and rmin */
  rminn = rmin / dr;
  rmaxn = rmax/ dr;

  if (WasDumpFileGiven) {
    Restore(theDumpFile);
  }
  else 
    load(initn);

  /**************************************************************/
  /* Deciding which circuit solver to use  */
 
  if (rmin > 0.0)
  {
   cent_cond = 1;
  for(i=0; i< ngr; i++)
   rg[i] =rminn +i;

  if(extl <1e-30 && extr <1e-30 &&
          (extc*log(rmaxn/rminn)/(TWOPI*zlength*epsilon)) >= 1e5)
    {    /* When L=R=0. and C tends to infinite */
      bcptr = bc3;
      circuitptr = circuit3;
      ntri=2; 
      if (BCflag==1) ntri--;
      b0 = 2*rg[1];
    }
  else  /* The general case with external voltage source */
    {
      bcptr = bc4;
      circuitptr = circuit4;

      ntri =1;

      b0= 1/(2.25*extl/dt/dt +1.5*extr/dt +1/extc);
      b0 = 1.+rminn*dr*b0/((rminn+0.5)*epsilon*area);
    }

  }

  fclose(InputDeck);
}   /* End START */

/***************************************************************/

int species(isp, initn, jj0)
int isp;
SCALAR *initn, jj0[];
{
  char aachar[80];
  SCALAR j0l, j0r, r0, r1, th0, th1;
  int   i;
  
  /* Read in SPECIES parameters... */
#ifndef XG_SCALAR_DOUBLE
  while (fscanf(InputDeck, "%g %g %g %g %d %d", 
#else
      while (fscanf(InputDeck, "%lg %lg %lg %lg %d %d",   //%d removed
#endif
      &q[isp], &m[isp], &enter[isp], &initn[isp], &sp_k[isp], &maxnp[isp]) <6)
/*    	&enter[isp], &initn[isp], &sp_k[isp], &maxnp[isp],&gyro_flag[isp]) <7)*/
    fscanf(InputDeck, "%s", aachar);
#ifndef XG_SCALAR_DOUBLE
while (fscanf(InputDeck, "%g %g %g %g %d %d", 
#else
    while (fscanf(InputDeck, "%lg %lg %lg %lg %d %d", 
#endif
      &v0x[isp], &v0y[isp], &v0z[isp], &vt[isp], &loader[isp], &inject_max_np[isp]) <6)
    fscanf(InputDeck, "%s", aachar);

  sr0[isp]=(SCALAR *)malloc(loader[isp]*sizeof(SCALAR));
  sth0[isp]=(SCALAR *)malloc(loader[isp]*sizeof(SCALAR));
  sr1[isp]=(SCALAR *)malloc(loader[isp]*sizeof(SCALAR));
  sth_width[isp]=(SCALAR *)malloc(loader[isp]*sizeof(SCALAR));

  for (i=0;i<loader[isp];i++) {
#ifndef XG_SCALAR_DOUBLE
      while(fscanf(InputDeck,"%g %g %g %g %g",
#else
          while(fscanf(InputDeck,"%lg %lg %lg %lg %lg",
#endif
        &r0, &th0, &r1, &th1, &r_width) < 5)
      fscanf(InputDeck, "%s", aachar);
    sr0[isp][i]=r0/dr; 
    sr1[isp][i]=r1/dr;
    r_width/=dr;
    sth0[isp][i]=th0*TWOPI/360.0/dth; 
    sth_width[isp][i]=(th1-th0)*TWOPI/360.0/dth;
  }
  /* Assign input parameters to arguments */

    jj0[isp] = fabs (j0l);

}   /* End SPECIES */

/***************************************************************/

int display_title()
{
  puts("\n\nPDC2.m - R, Theta Electrostatic 2 Dimensional Code");
  puts(" Version 2.0");
  puts(" (c) Copyright 1992 Regents of the University of California");
  puts("Plasma Theory and Simulation Group");
  puts("Originally developed by J. P. Gopinath (1992)");
  puts("Modified by Hae June Lee (2001).");
}
/***************************************************************/

int PeriodicDump()
{
  int i;

  DropFile = fopen(dropfile, "a");
  fprintf(DropFile, "%g ", t);
  for(i=0; i<nsp; i++) fprintf(DropFile, "%d ", np[i]);
  fprintf(DropFile, "\n");
  fclose(DropFile);
  Dump(dumpfile);
}

/***************************************************************/
